#include <cassert>
#include <iomanip>

#include "CoinPragma.hpp"
#include "CoinWarmStart.hpp"
#include "CoinWarmStartBasis.hpp"
#include "OsiClpSolverInterface.hpp"

void checkDualObjValue(OsiClpSolverInterface si, const double *pi = NULL, const double *dj = NULL){
  int numCols = si.getNumCols(); // n
  int numRows = si.getNumRows(); // m
  double dual_obj_val = 0;
  if (!pi)
    pi = si.getRowPrice();
  if (!dj)
    dj = si.getReducedCost();

  // pi * b
  for (int i = 0; i < numRows; i++){
    if (si.getRowUpper()[i] < 1000000)
      dual_obj_val += si.getRowUpper()[i] * pi[i];
    else
      dual_obj_val += si.getRowLower()[i] * pi[i];
  }
  
  // dj * lb/ub
  for (int i = 0; i < numCols; i++)
  {
    if (dj[i] < -1e-7)
        dual_obj_val += si.getColUpper()[i] * dj[i];
    else if (dj[i] > 1e-7)
        dual_obj_val += si.getColLower()[i] * dj[i];
  }

  if (si.isProvenOptimal())
    printf("CLP: isProvenOptimal\n");
  else if (si.isIterationLimitReached())
    printf("CLP: isIterationLimitReached\n");
  else if (si.isProvenPrimalInfeasible())
    printf("CLP: isProvenPrimalInfeasible\n");

  printf("Dual objective: %.10f\n", dual_obj_val);
  printf(" getObjValue(): %.10f\n", si.getObjValue());

  if (fabs(dual_obj_val - si.getObjValue()) <= 0.0001)
    printf("Values matches!\n");
  else
    printf("WARNING: Values mismatches!\n");
    printf("Gap: %.5f\n", fabs(dual_obj_val - si.getObjValue()));

}

bool checkDualSolution(OsiClpSolverInterface si){
  const double *c  = si.getObjCoefficients();
  const double *pi = si.getRowPrice();
  const double *dj = si.getReducedCost();
  const CoinPackedMatrix *A = si.getMatrixByCol();
  int n = si.getNumCols();
  int m = si.getNumRows();
  double *new_dj = new double[n];
  double *piA = new double[n];
  bool isDualFeas = true;
  int *basics = new int[m];
  CoinWarmStartBasis *basis = si.getModelPtr()->getBasis();
  
  A->transposeTimes(pi, piA);
  int j = 0;
  for (j = 0; j < n; j++){
    if (si.getColUpper()[j] == si.getColLower()[j])
      new_dj[j] = c[j] - piA[j];
    else
      new_dj[j] = dj[j];
    // printf("  Var %d : Status %d : dj %.4f : new_dj : %.4f\n", j, basis->getStructStatus(j), dj[j], new_dj[j]);
    if (piA[j] + dj[j] - c[j] > 1e-5){
      isDualFeas = false;
    }
  }

  if (isDualFeas){
    printf("Dual feasible!\n");
  } else {
    printf("\nWARNING: Dual infeasible!\n");
  }

  printf("Recomputing dj's...\n");
  checkDualObjValue(si, pi, new_dj);

  delete[] piA;
  delete[] new_dj;
  delete[] basics;
  return isDualFeas;
}


int main(int argc, const char *argv[])
{
  OsiClpSolverInterface si;
  int numberIterations, numCols, numRows;

  std::string mpsFileName;
  if (argc >= 2) {
    mpsFileName = argv[1];

    int numMpsReadErrors = si.readMps(mpsFileName.c_str(), "");
   
    if (numMpsReadErrors != 0) {
      printf("%d errors reading MPS file\n", numMpsReadErrors);
      return numMpsReadErrors;
    }
  } 

  // Setup OsiClp options
  si.setupForRepeatedUse(2, 0);
  int options = si.getModelPtr()->specialOptions();
  options |= 32 + 0x08000000;
  si.getModelPtr()->setSpecialOptions(options);
  si.getModelPtr()->setPerturbation(50);
  options = si.getModelPtr()->specialOptions();
  si.setCleanupScaling(1);

  si.getModelPtr()->setMoreSpecialOptions(si.getModelPtr()->moreSpecialOptions()|8192);

  // Only dual simplex should be used and no presolve
  // si.setHintParam(OsiDoReducePrint);
  si.setLogLevel(63);

  si.setHintParam(OsiDoScale,false,OsiHintDo);
  si.setHintParam(OsiDoPresolveInInitial, false, OsiHintDo);
  si.setHintParam(OsiDoPresolveInResolve, false, OsiHintDo);
  si.setHintParam(OsiDoDualInInitial, true, OsiHintDo);
  si.setHintParam(OsiDoDualInResolve, true, OsiHintDo);

  numCols = si.getNumCols(); // n
  numRows = si.getNumRows(); // m

  // Solve the "root node" LP
  printf("Solving the root LP...\n");
  si.initialSolve();
  printf("initialSolve()\n");
  numberIterations = si.getIterationCount();

  checkDualObjValue(si);
  checkDualSolution(si);
  printf("\n");

  if (si.isProvenPrimalInfeasible()){
    double *ray = si.getDualRays(1, false)[0];
    double *rayA = (double *)malloc(sizeof(double) * numCols);

    const CoinPackedMatrix *A = si.getMatrixByCol();
         
    A->transposeTimes(ray, rayA);
    double farkasProof = 0;

    for (int i = 0; i < numRows; i++){
        farkasProof -= si.getRightHandSide()[i] * ray[i];
    }

    for (int j = 0; j < numCols; j++){
        if (rayA[j] > 1e-5){
          farkasProof += si.getColLower()[j] * rayA[j];
        } else if (rayA[j] < -1e-5) {
          farkasProof += si.getColUpper()[j] * rayA[j];
        }
    }

    if (farkasProof < -1e-5){
        printf("Farkas proof: rAx - rb > 0 fails!\n");
        printf("rAx - rb: %.10f\n", farkasProof);
    }

    OsiClpSolverInterface solver(si);

    int *index = new int[CoinMax(numCols, numRows)];
    int i;
    for (i = CoinMax(numCols, numRows) - 1; i >= 0; --i) {
      index[i] = i;
    }
    double *obj = new double[CoinMax(numCols, 2 * numRows)];
    CoinFillN(obj, numCols, 0.0);
    solver.setObjCoeffSet(index, index + numCols, obj);

    double *clb = new double[2 * numRows];
    double *cub = new double[2 * numRows];

    const double plusone = 1.0;
    const double minusone = -1.0;
    const char *sense = si.getRowSense();

    const CoinPackedVectorBase **cols = new const CoinPackedVectorBase *[numRows];
    int newcols = 0;
    for (i = 0; i < numRows; ++i) {
      switch (sense[i]) {
      case 'L':
        cols[newcols++] = new CoinShallowPackedVector(1, &index[i], &minusone, false);
        break;
      case 'G':
        cols[newcols++] = new CoinShallowPackedVector(1, &index[i], &plusone, false);
        break;
      case 'R':
        cols[newcols++] = new CoinShallowPackedVector(1, &index[i], &minusone, false);
        cols[newcols++] = new CoinShallowPackedVector(1, &index[i], &plusone, false);
        break;
      case 'N':
      case 'E':
        break;
      }
    }

    CoinFillN(obj, newcols, 1.0);
    CoinFillN(clb, newcols, 0.0);
    CoinFillN(cub, newcols, si.getInfinity());

    solver.addCols(newcols, cols, clb, cub, obj);
    delete[] index;
    delete[] cols;
    delete[] clb;
    delete[] cub;
    delete[] obj;

    solver.setObjSense(1.0); // minimize
    solver.initialSolve();

    const double *solverpi = solver.getRowPrice();
    double *lambda = new double[numRows];
    for (i = numRows - 1; i >= 0; --i) {
      lambda[i] = -solverpi[i];
    }

    ray = lambda;

    A->transposeTimes(ray, rayA);
    farkasProof = 0;

    for (int i = 0; i < numRows; i++){
        farkasProof -= si.getRightHandSide()[i] * ray[i];
    }

    for (int j = 0; j < numCols; j++){
        if (rayA[j] > 1e-5){
          farkasProof += si.getColLower()[j] * rayA[j];
        } else if (rayA[j] < -1e-5) {
          farkasProof += si.getColUpper()[j] * rayA[j];
        }
    }

    if (farkasProof < -1e-5){
        printf("Farkas proof: rAx - rb > 0 fails!\n");
        printf("rAx - rb: %.10f\n", farkasProof);
    }
    
    delete[] ray;
    free(rayA); 
      
  }


  return 0;
}
